//--------------------------------------------------------------------------------------
// File: Prototype4.cpp
//
// (FreeBSD License)
// 
// Copyright 2010 Andreas Kirsch. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are
// permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice, this list of
// conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list
// of conditions and the following disclaimer in the documentation and/or other materials
// provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY ANDREAS KIRSCH ``AS IS'' AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ANDREAS KIRSCH OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
// ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// The views and conclusions contained in the software and documentation are those of the
// authors and should not be interpreted as representing official policies, either expressed
// or implied, of Andreas Kirsch.
//--------------------------------------------------------------------------------------
#include <math.h>
#include <algorithm>  

#include <boost/lexical_cast.hpp>

#include "DXUT.h"
#include "DXUTgui.h"
#include "DXUTmisc.h"
#include "DXUTCamera.h"
#include "DXUTSettingsDlg.h"
#include "SDKmisc.h"
#include "SDKmesh.h"
#include "resource.h"
#include "DXUTShapes.h"

#include "config.h"

#include "material.h"
#include "dxhelpers.h"
#include "reflectiveShadowMap.h"
#include "shadowMap.h"
#include "MeshLoader10.h"
#include "GPUlpv.h"

const wchar_t *gc_dumpName = L"sponza.mesh"; 
const wchar_t *gc_modelFilename = L"..\\Media\\CrySponza\\sponza-triangulated-c.obj"; 
const float gc_zNear = 0.1f;
const float gc_zFar = 1000.0f;

struct Light {
	LPV::ReflectiveShadowMap rsm;
	LPV::ShadowMap sm;

	const D3DXVECTOR3 & position() const {
		return *camera.GetEyePt();
	}

	const D3DXVECTOR3 & lookAt() const {
		return *camera.GetLookAtPt();
	}

	D3DXMATRIX worldViewProjMatrix;

	D3DXVECTOR3 flux;

	CFirstPersonCamera camera;
};

struct RenderContext {
	D3DXMATRIX projectionMatrix, viewMatrix;
	D3DXMATRIX worldMatrix;

	bool rsmPass;
	Light *activeLight;

	ID3D10EffectTechnique *technique;
};

LPV::GPULightPropagationVolume *lpv;

void initScene( ID3D10Device *device );
void destroyScene();
void renderScene(  ID3D10Device *device, const RenderContext &renderContext );
void renderSceneFromCamera( ID3D10Device *device, const CBaseCamera *camera);
void renderSM( ID3D10Device *device, Light &light );
void renderRSM( ID3D10Device *device, Light &light );

//#define DEBUG_VS   // Uncomment this line to debug D3D9 vertex shaders 
//#define DEBUG_PS   // Uncomment this line to debug D3D9 pixel shaders 


//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------

std::vector< CBaseCamera * > g_RegisteredCameras;

CBaseCamera*				g_ActiveCamera;
CFirstPersonCamera          g_SceneCamera;               // A model viewing camera
CDXUTDialogResourceManager  g_DialogResourceManager; // manager for shared resources of dialogs
CD3DSettingsDlg             g_SettingsDlg;          // Device settings dialog
CDXUTTextHelper*            g_pTxtHelper = NULL;
CDXUTDialog                 g_HUD;                  // dialog for standard controls
CDXUTDialog                 g_SampleUI;             // dialog for sample specific controls

CMeshLoader10               g_MeshLoader;            // Loads a mesh from an .obj fil
bool g_bReloadEffect = false;
bool g_bRestoreTriggered = false;
bool g_bReloadInjectionFX = false;
int g_iLPVIndex = -1;
bool g_bRenderDirectLighting = true;

const int g_fIndirectLightingAttenuation_UnitSteps = 4;
const int g_fIndirectLightingAttenuation_Max = 5;
float g_fIndirectLightingAttenuation = 1.0f;

// Direct3D 10 resources
ID3DX10Font*                g_pFont10 = NULL;
ID3DX10Sprite*              g_pSprite10 = NULL;

ID3D10Effect*                       g_pEffect10 = NULL;
ID3D10InputLayout*                  g_pVertexLayout = NULL;
ID3D10EffectTechnique*              g_pRenderTechnique = NULL;

ID3D10EffectTechnique*              g_pRenderLVTechnique = NULL;
ID3D10EffectTechnique*              g_pRenderRCTechnique = NULL;
ID3D10EffectTechnique*              g_pRSMTechnique = NULL;
ID3D10EffectTechnique*              g_pSMTechnique = NULL;

ID3D10EffectScalarVariable*         g_pRSMResolution = NULL;

ID3D10EffectShaderResourceVariable* g_ptxDiffuseVariable = NULL;

ID3D10EffectVectorVariable*         g_pAmbient = NULL;
ID3D10EffectVectorVariable*         g_pDiffuse = NULL;
ID3D10EffectVectorVariable*         g_pSpecular = NULL;
ID3D10EffectScalarVariable*         g_pOpacity = NULL;
ID3D10EffectScalarVariable*         g_pSpecularPower = NULL;

ID3D10EffectVectorVariable*         g_pLightColor = NULL;
ID3D10EffectVectorVariable*         g_pLightPosition = NULL;
ID3D10EffectShaderResourceVariable* g_pLightDepthMap = NULL;
ID3D10EffectMatrixVariable*			g_pLightProjMatrix = NULL;

ID3D10EffectScalarVariable*			g_pRenderDirectLighting = NULL;
ID3D10EffectScalarVariable*			g_pIndirectLightingAttenuation = NULL;

ID3D10EffectVectorVariable*         g_pCameraPosition = NULL;
ID3D10EffectScalarVariable*         g_pTime = NULL;
ID3D10EffectMatrixVariable*         g_pProjection = NULL;
ID3D10EffectMatrixVariable*         g_pWorld = NULL;
ID3D10EffectMatrixVariable*         g_pView = NULL;
ID3D10EffectMatrixVariable*         g_pInvWorld = NULL;
ID3D10EffectMatrixVariable*         g_pInvView = NULL;
ID3D10EffectMatrixVariable*         g_pWorldViewProjection = NULL;

ID3D10EffectShaderResourceVariable* g_pLV = NULL;
ID3D10EffectShaderResourceVariable* g_pGV = NULL;
ID3D10EffectVectorVariable*			g_pLV_MinCorner = NULL;
ID3D10EffectScalarVariable*			g_pLV_CellSize = NULL;

ID3D10Texture2D* g_whiteDummyTexture = NULL;
ID3D10ShaderResourceView* g_whiteDummyTextureSRV = NULL;

Light g_Light;

HRESULT createWhiteDummyTexture( ID3D10Device *device ) {
	HRESULT hr;

	static const int dummyData[1] = { ~0 };

	V_RETURN( device->CreateTexture2D( D3D10Texture2DDesc( DXGI_FORMAT_R8G8B8A8_UNORM, D3D10_USAGE_IMMUTABLE, 1, 1 ).noMipMapLevels().bindFlags( D3D10_BIND_FLAG_CPP().useAsShaderResource() ), D3D10SubResourceData( &dummyData ).sysMemPitch( 4 ), &g_whiteDummyTexture ) );
	V_RETURN( device->CreateShaderResourceView( g_whiteDummyTexture, NULL, &g_whiteDummyTextureSRV ) );

	return S_OK;
}


//--------------------------------------------------------------------------------------
// UI control IDs
//--------------------------------------------------------------------------------------
#define IDC_TOGGLEFULLSCREEN    1
#define IDC_TOGGLEREF           2
#define IDC_CHANGEDEVICE        3
#define IDC_TOGGLEWARP          4

#define IDC_SWITCH_CAMERA		5

#define IDC_RELOAD_EFFECT		6
#define IDC_DUMP				7
#define IDC_RESTORE				8
#define IDC_SUBSET              9
#define IDC_RELOAD_INJECTION_EFFECT		10
#define IDC_LPV_CHOOSER		11
#define IDC_OCCLUSION_SWITCH		12
#define IDC_DIRECT_LIGHTING_SWITCH 13
#define IDC_GI_BOOST_VALUE	14
#define IDC_GI_BOOST_INFO	15

#define IDC_STATIC				255
//--------------------------------------------------------------------------------------
// Forward declarations 
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing,
                          void* pUserContext );
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext );
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext );
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext );
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext );

bool CALLBACK IsD3D10DeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType,
                                       DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext );
HRESULT CALLBACK OnD3D10CreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc,
                                      void* pUserContext );
HRESULT CALLBACK OnD3D10ResizedSwapChain( ID3D10Device* pd3dDevice, IDXGISwapChain* pSwapChain,
                                          const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
void CALLBACK OnD3D10FrameRender( ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext );
void CALLBACK OnD3D10ReleasingSwapChain( void* pUserContext );
void CALLBACK OnD3D10DestroyDevice( void* pUserContext );

void InitApp();
void RenderText();


//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------
int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
    // Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
    _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

    // DXUT will create and use the best device (either D3D9 or D3D10) 
    // that is available on the system depending on which D3D callbacks are set below

    // Set DXUT callbacks
    DXUTSetCallbackMsgProc( MsgProc );
    DXUTSetCallbackKeyboard( OnKeyboard );
    DXUTSetCallbackFrameMove( OnFrameMove );
    DXUTSetCallbackDeviceChanging( ModifyDeviceSettings );

    DXUTSetCallbackD3D10DeviceAcceptable( IsD3D10DeviceAcceptable );
    DXUTSetCallbackD3D10DeviceCreated( OnD3D10CreateDevice );
    DXUTSetCallbackD3D10SwapChainResized( OnD3D10ResizedSwapChain );
    DXUTSetCallbackD3D10SwapChainReleasing( OnD3D10ReleasingSwapChain );
    DXUTSetCallbackD3D10DeviceDestroyed( OnD3D10DestroyDevice );
    DXUTSetCallbackD3D10FrameRender( OnD3D10FrameRender );

    InitApp();
    DXUTInit( true, true, NULL ); // Parse the command line, show msgboxes on error, no extra command line params
    DXUTSetCursorSettings( true, true );
    DXUTCreateWindow( L"LPVPrototype" );
    DXUTCreateDevice( true, 640, 480 );
    DXUTMainLoop(); // Enter into the DXUT render loop

    return DXUTGetExitCode();
}

//////////////////////////////////////////////////////////////////////////

void destroyEffect() {
	SAFE_RELEASE(g_pVertexLayout );
	SAFE_RELEASE( g_pEffect10 );
}

HRESULT loadEffect( ID3D10Device *device ) {
	HRESULT hr;

	// Read the D3DX effect file
	WCHAR str[MAX_PATH];
	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"scene.fx" ) );
	DWORD dwShaderFlags = D3D10_SHADER_ENABLE_STRICTNESS;
#if defined( DEBUG ) || defined( _DEBUG )
	// Set the D3D10_SHADER_DEBUG flag to embed debug information in the shaders.
	// Setting this flag improves the shader debugging experience, but still allows
	// the shaders to be optimized and to run exactly the way they will run in
	// the release configuration of this program.
	dwShaderFlags |= D3D10_SHADER_DEBUG;
#endif
	V_RETURN( D3DX10CreateEffectFromFile( str, NULL, NULL, "fx_4_0", dwShaderFlags, 0, device, NULL,
		NULL, &g_pEffect10, NULL, NULL ) );

	// Obtain the technique
	g_pRenderTechnique = g_pEffect10->GetTechniqueByName( "UnifiedRendering" );
	g_pRenderLVTechnique = g_pEffect10->GetTechniqueByName( "UnifiedLVRendering" );
	g_pRenderRCTechnique = g_pEffect10->GetTechniqueByName( "UnifiedRCRendering" );
	g_pRSMTechnique = g_pEffect10->GetTechniqueByName( "RSMRendering" );
	g_pSMTechnique = g_pEffect10->GetTechniqueByName( "SMRendering" );

	g_pRSMResolution = g_pEffect10->GetVariableByName( "g_RSMResolution" )->AsScalar();    

	g_ptxDiffuseVariable = g_pEffect10->GetVariableByName( "g_MeshTexture" )->AsShaderResource();    

	g_pAmbient = g_pEffect10->GetVariableByName( "g_vMaterialAmbient" )->AsVector();
	g_pDiffuse = g_pEffect10->GetVariableByName( "g_vMaterialDiffuse" )->AsVector();
	g_pSpecular = g_pEffect10->GetVariableByName( "g_vMaterialSpecular" )->AsVector();
	g_pOpacity = g_pEffect10->GetVariableByName( "g_fMaterialAlpha" )->AsScalar();
	g_pSpecularPower = g_pEffect10->GetVariableByName( "g_nMaterialShininess" )->AsScalar();

	g_pLightColor = g_pEffect10->GetVariableByName( "g_vLightColor" )->AsVector();
	g_pLightPosition = g_pEffect10->GetVariableByName( "g_vLightPosition" )->AsVector();

	g_pLightDepthMap = g_pEffect10->GetVariableByName( "g_LightDepthMap" )->AsShaderResource();
	g_pLightProjMatrix = g_pEffect10->GetVariableByName( "g_mLightProjMatrix" )->AsMatrix();

	g_pRenderDirectLighting = g_pEffect10->GetVariableByName( "g_renderDirectLighting" )->AsScalar();
	g_pIndirectLightingAttenuation = g_pEffect10->GetVariableByName( "g_indirectLightingAttenuation" )->AsScalar();

	g_pCameraPosition = g_pEffect10->GetVariableByName( "g_vCameraPosition" )->AsVector();
	g_pTime = g_pEffect10->GetVariableByName( "g_fTime" )->AsScalar();
	g_pProjection = g_pEffect10->GetVariableByName( "g_mProjection" )->AsMatrix();
	g_pWorld = g_pEffect10->GetVariableByName( "g_mWorld" )->AsMatrix();
	g_pView = g_pEffect10->GetVariableByName( "g_mView" )->AsMatrix();
	g_pInvWorld = g_pEffect10->GetVariableByName( "g_mInvWorld" )->AsMatrix();
	g_pInvView = g_pEffect10->GetVariableByName( "g_mInvView" )->AsMatrix();
	g_pWorldViewProjection = g_pEffect10->GetVariableByName( "g_mWorldViewProjection" )->AsMatrix();

	g_pLV = g_pEffect10->GetVariableByName( "g_LV" )->AsShaderResource();
	g_pGV = g_pEffect10->GetVariableByName( "g_GV" )->AsShaderResource();

	g_pLV_MinCorner = g_pEffect10->GetVariableByName( "g_LV_MinCorner" )->AsVector();
	g_pLV_CellSize = g_pEffect10->GetVariableByName( "g_LV_CellSize" )->AsScalar();

	// Define the input layout
	const D3D10_INPUT_ELEMENT_DESC layout[] =
	{
		{ "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D10_INPUT_PER_VERTEX_DATA, 0 },
		{ "NORMAL", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D10_INPUT_PER_VERTEX_DATA, 0 },
		{ "TEXCOORD", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 24, D3D10_INPUT_PER_VERTEX_DATA, 0 },
	} ;
	UINT numElements = sizeof( layout ) / sizeof( layout[0] );

	// Create the input layout
	D3D10_PASS_DESC PassDesc;
	g_pRenderTechnique->GetPassByIndex( 0 )->GetDesc( &PassDesc );
	V_RETURN( device->CreateInputLayout( layout, numElements, PassDesc.pIAInputSignature,
		PassDesc.IAInputSignatureSize, &g_pVertexLayout ) );

	device->IASetInputLayout( g_pVertexLayout );

	return S_OK;
}

void reloadEffect( ID3D10Device *device ) {
	destroyEffect();

	loadEffect( device );
}

//--------------------------------------------------------------------------------------

void Effect_setMaterial( const Material &material ) {
	HRESULT hr;

	V( g_pAmbient->SetFloatVector( (float*) (const float*) material.vAmbient ) );
	V( g_pDiffuse->SetFloatVector( (float*) (const float*) material.vDiffuse ) );
	if( material.bSpecular ) {
		V( g_pSpecular->SetFloatVector( (float*) (const float*) material.vSpecular ) );
	}
	else {
		V( g_pSpecular->SetFloatVector( D3DXVECTOR3( 0.0f, 0.0f, 0.0f ) ) );
	}
	V( g_pOpacity->SetFloat( material.fAlpha ) );
	V( g_pSpecularPower->SetInt( material.nShininess ) );

	if ( !IsErrorResource( material.pTextureRV10 ) && material.pTextureRV10 ) {
		g_ptxDiffuseVariable->SetResource( material.pTextureRV10 );
	}
	else {
		g_ptxDiffuseVariable->SetResource( g_whiteDummyTextureSRV );
	}
}

void Effect_setLight( const Light &light, bool shadowMap = true ) {
	g_pLightPosition->SetFloatVector( (float*) (const float*) light.position() );
	g_pLightColor->SetFloatVector( (float*) (const float*) light.flux );
	g_pLightProjMatrix->SetMatrix( (float*) (const float*) light.worldViewProjMatrix );
	if( shadowMap ) {
		g_pLightDepthMap->SetResource( light.sm.shadowMapSRV );
	}
	else {
		g_pLightDepthMap->SetResource( NULL );
	}
}

void Effect_setTransformation( const D3DXMATRIX &world, const D3DXMATRIX &view, const D3DXMATRIX &projection ) {
	g_pProjection->SetMatrix( (float*) (const float*) projection );

	g_pWorld->SetMatrix( (float*) (const float*) world );
	g_pView->SetMatrix( (float*) (const float*) view );
	
	D3DXMATRIX invView, invWorld;
	D3DXMatrixInverse( &invWorld, NULL, &world );
	D3DXMatrixInverse( &invView, NULL, &view );
	g_pInvWorld->SetMatrix( (float*) (const float*) invWorld );
	g_pInvView->SetMatrix( (float*) (const float*) invView );

	D3DXMATRIX worldViewProjection = world * view * projection;
	g_pWorldViewProjection->SetMatrix( (float*) (const float*) worldViewProjection );
}

void Effect_setCameraPosition( const D3DXVECTOR3 &position ) {
	g_pCameraPosition->SetFloatVector( (float*) (const float*) position );
}

void renderSubset( const RenderContext &context, UINT iSubset )
{
	Material* pMaterial = g_MeshLoader.GetSubsetMaterial( iSubset );

	Effect_setMaterial( *pMaterial );

	D3D10_TECHNIQUE_DESC techDesc;
	context.technique->GetDesc( &techDesc );

	for ( UINT p = 0; p < techDesc.Passes; ++p )
	{
		context.technique->GetPassByIndex(p)->Apply(0);
		g_MeshLoader.GetMesh()->DrawSubset(iSubset);
	}
}

//////////////////////////////////////////////////////////////////////////


//--------------------------------------------------------------------------------------
// Initialize the app 
//--------------------------------------------------------------------------------------
void InitApp()
{
    g_SettingsDlg.Init( &g_DialogResourceManager );
    g_HUD.Init( &g_DialogResourceManager );
    g_SampleUI.Init( &g_DialogResourceManager );

    g_HUD.SetCallback( OnGUIEvent ); int iY = 10;
    g_HUD.AddButton( IDC_TOGGLEFULLSCREEN, L"Toggle full screen", 35, iY, 125, 22 );
    g_HUD.AddButton( IDC_CHANGEDEVICE, L"Change device (F2)", 35, iY += 24, 125, 22, VK_F2 );
    g_HUD.AddButton( IDC_TOGGLEREF, L"Toggle REF (F3)", 35, iY += 24, 125, 22, VK_F3 );
    g_HUD.AddButton( IDC_TOGGLEWARP, L"Toggle WARP (F4)", 35, iY += 24, 125, 22, VK_F4 );
	g_HUD.AddButton( IDC_RELOAD_EFFECT, L"Reload scene.fx (F5)", 35, iY += 24, 125, 22, VK_F5 );

    g_SampleUI.SetCallback( OnGUIEvent ); iY = 10;
	g_SampleUI.AddButton( IDC_SWITCH_CAMERA, L"Switch Camera", 35, iY += 24, 125, 22 );

	g_RegisteredCameras.push_back( &g_SceneCamera );
	g_ActiveCamera = &g_SceneCamera;

	//g_SampleUI.AddButton( IDC_DUMP, L"Dump (F6)", 35, iY += 24, 125, 22, VK_F6 );
	//g_SampleUI.AddButton( IDC_RESTORE, L"Restore", 35, iY += 24, 125, 22 );

	// Title font for comboboxes
	g_SampleUI.SetFont( 1, L"Arial", 14, FW_BOLD );
	CDXUTElement* pElement = g_SampleUI.GetDefaultElement( DXUT_CONTROL_STATIC, 0 );
	if( pElement )
	{
		pElement->iFont = 1;
		pElement->dwTextFormat = DT_LEFT | DT_BOTTOM;
	}

	g_SampleUI.AddStatic( IDC_STATIC, L"(S)ubset", 20, iY += 24, 105, 25 );
	g_SampleUI.AddComboBox( IDC_SUBSET, 20, iY += 24, 140, 24 );

	g_SampleUI.AddButton( IDC_RELOAD_INJECTION_EFFECT, L"Reload lpv.fx", 35, iY += 24, 125, 22 );

	g_SampleUI.AddStatic( IDC_STATIC, L"LPV Volume:", 35, iY += 24, 70, 22 );
	g_SampleUI.AddSlider( IDC_LPV_CHOOSER, 50, iY += 24 , 80, 22, -1, LPV::GPULightPropagationVolume::LPV_COUNT - 1, g_iLPVIndex );
	
	g_SampleUI.AddCheckBox( IDC_OCCLUSION_SWITCH, L"Occlusion", 35, iY += 24, 105, 20, true );

	g_SampleUI.AddCheckBox( IDC_DIRECT_LIGHTING_SWITCH, L"Direct Lighting", 35, iY += 24, 105, 20, true );
	
	g_SampleUI.AddStatic( IDC_STATIC, L"GI Boost:", 35, iY += 24, 105, 22 );
	g_SampleUI.AddSlider( IDC_GI_BOOST_VALUE, 50, iY += 24, 80, 22, 0, g_fIndirectLightingAttenuation_Max * g_fIndirectLightingAttenuation_UnitSteps, g_fIndirectLightingAttenuation * g_fIndirectLightingAttenuation_UnitSteps );

	g_SampleUI.AddStatic( IDC_STATIC, L"  Value:", 50, iY += 24, 105, 22 );
	g_SampleUI.AddStatic( IDC_GI_BOOST_INFO, boost::lexical_cast<std::wstring>(g_fIndirectLightingAttenuation).c_str(), 90, iY, 80, 22 );
}

//--------------------------------------------------------------------------------------
// Render the help and statistics text. This function uses the ID3DXFont interface for 
// efficient text rendering.
//--------------------------------------------------------------------------------------
void RenderText()
{
    g_pTxtHelper->Begin();
    g_pTxtHelper->SetInsertionPos( 5, 5 );
    g_pTxtHelper->SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 0.0f, 1.0f ) );
    g_pTxtHelper->DrawTextLine( DXUTGetFrameStats( DXUTIsVsyncEnabled() ) );
    g_pTxtHelper->DrawTextLine( DXUTGetDeviceStats() );
    g_pTxtHelper->End();
}


//--------------------------------------------------------------------------------------
// Reject any D3D10 devices that aren't acceptable by returning false
//--------------------------------------------------------------------------------------
bool CALLBACK IsD3D10DeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType,
                                       DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext )
{
    return true;
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that aren't dependant on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10CreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc,
                                      void* pUserContext )
{
    HRESULT hr;

    V_RETURN( D3DX10CreateSprite( pd3dDevice, 500, &g_pSprite10 ) );
    V_RETURN( g_DialogResourceManager.OnD3D10CreateDevice( pd3dDevice ) );
    V_RETURN( g_SettingsDlg.OnD3D10CreateDevice( pd3dDevice ) );
    V_RETURN( D3DX10CreateFont( pd3dDevice, 15, 0, FW_BOLD, 1, FALSE, DEFAULT_CHARSET,
                                OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE,
                                L"Arial", &g_pFont10 ) );
    g_pTxtHelper = new CDXUTTextHelper( NULL, NULL, g_pFont10, g_pSprite10, 15 );

	V( createWhiteDummyTexture( pd3dDevice ) );

	// Load the mesh
	//V_RETURN( g_MeshLoader.Create( pd3dDevice, gc_modelFilename, 100.0f ) );
	V_RETURN( g_MeshLoader.Restore( pd3dDevice, gc_dumpName ) );

	lpv = new LPV::GPULightPropagationVolume();
	lpv->init( pd3dDevice, LPV_RESOLUTION, g_MeshLoader.GetMinCorner(), g_MeshLoader.GetMaxCorner() );

	V_RETURN( loadEffect( pd3dDevice ) );

	// Add the identified subsets to the UI
	CDXUTComboBox* pComboBox = g_SampleUI.GetComboBox( IDC_SUBSET );
	pComboBox->RemoveAllItems();
	pComboBox->AddItem( L"All", ( void* )( INT_PTR )-1 );

	for ( UINT subset = 0; subset < g_MeshLoader.GetNumSubsets(); ++subset )
	{
		Material* pMaterial = g_MeshLoader.GetSubsetMaterial( subset );
		pComboBox->AddItem( pMaterial->strName, ( void* )( INT_PTR )subset );
	}    

    // Setup the camera's view parameters
    D3DXVECTOR3 vecEye( 0.0f, 3.0f, -5.0f );
    D3DXVECTOR3 vecAt ( 0.0f, 0.0f, -0.0f );
    g_SceneCamera.SetViewParams( &vecEye, &vecAt );

	initScene( pd3dDevice );

    return S_OK;
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that depend on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10ResizedSwapChain( ID3D10Device* pd3dDevice, IDXGISwapChain* pSwapChain,
                                          const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr;

    V_RETURN( g_DialogResourceManager.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );
    V_RETURN( g_SettingsDlg.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );

    // Setup the camera's projection parameters
    float fAspectRatio = pBackBufferSurfaceDesc->Width / ( FLOAT )pBackBufferSurfaceDesc->Height;
    g_SceneCamera.SetProjParams( D3DX_PI / 4, fAspectRatio, gc_zNear, gc_zFar );
    //g_SceneCamera.SetWindow( pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height );
    //g_SceneCamera.SetButtonMasks( MOUSE_LEFT_BUTTON, MOUSE_WHEEL, MOUSE_MIDDLE_BUTTON );

    g_HUD.SetLocation( pBackBufferSurfaceDesc->Width - 170, 0 );
    g_HUD.SetSize( 170, 170 );
    g_SampleUI.SetLocation( pBackBufferSurfaceDesc->Width - 170, pBackBufferSurfaceDesc->Height - 380 );
    g_SampleUI.SetSize( 170, 380 );

    return S_OK;
}


//--------------------------------------------------------------------------------------
// Render the scene using the D3D10 device
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10FrameRender( ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
	HRESULT hr;

	if( g_bReloadEffect ) {
		g_bReloadEffect = false;
		reloadEffect( pd3dDevice );
	}
	if( g_bReloadInjectionFX ) {
		g_bReloadInjectionFX = false;
		lpv->reloadEffect( pd3dDevice );
	}

	if( g_bRestoreTriggered ) {
		g_bRestoreTriggered = false;
		g_MeshLoader.Restore( pd3dDevice, gc_dumpName );
	}

	g_pRenderDirectLighting->SetBool( g_bRenderDirectLighting );
	g_pIndirectLightingAttenuation->SetFloat( g_fIndirectLightingAttenuation );

    float ClearColor[4] = { 0.176f, 0.196f, 0.667f, 0.0f };
    ID3D10RenderTargetView* pRTV = DXUTGetD3D10RenderTargetView();
    pd3dDevice->ClearRenderTargetView( pRTV, ClearColor );

    // Clear the depth stencil
    ID3D10DepthStencilView* pDSV = DXUTGetD3D10DepthStencilView();
    pd3dDevice->ClearDepthStencilView( pDSV, D3D10_CLEAR_DEPTH, 1.0, 0 );

    // If the settings dialog is being shown, then render it instead of rendering the app's scene
    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.OnRender( fElapsedTime );
        return;
    }


    // Update the effect's variables.  Instead of using strings, it would 
    // be more efficient to cache a handle to the parameter by calling 
    // ID3DXEffect::GetParameterByName
//     g_pmWorldViewProj->SetMatrix( ( float* )&mWorldViewProjection );
//     g_pmWorld->SetMatrix( ( float* )&mWorld );
	V( g_pTime->SetFloat( (float)fTime ) );

	DXUT_BeginPerfEvent( DXUT_PERFEVENTCOLOR, L"Render SM" );
	renderSM( pd3dDevice, g_Light ); 
	DXUT_EndPerfEvent();

	DXUT_BeginPerfEvent( DXUT_PERFEVENTCOLOR, L"Render RSM" );
	renderRSM( pd3dDevice, g_Light );
	DXUT_EndPerfEvent();
		
	DXUT_BeginPerfEvent( DXUT_PERFEVENTCOLOR, L"Inject RSM" );

	ID3D10ShaderResourceView *nullView[4] = { NULL };
	g_pLV->SetResourceArray( nullView, 0, 4 );
	lpv->injectRSM( pd3dDevice, g_Light.rsm );

	DXUT_EndPerfEvent();

	DXUT_BeginPerfEvent( DXUT_PERFEVENTCOLOR, L"Propagate Light" );

	lpv->propagate( pd3dDevice );

	DXUT_EndPerfEvent();

	pd3dDevice->OMSetRenderTargets( 1, &pRTV, pDSV );

	// TODO: move this into its own function! [6/6/2010 Andreas Kirsch]
	if( g_iLPVIndex == -1 ) {
		lpv->accumulatedLPV.setResource( g_pLV );
	}
	else {
		lpv->LPVs[ g_iLPVIndex ].setResource( g_pLV );
	}
	
	lpv->geometryVolume.setResource( g_pGV );
	
	g_pLV_MinCorner->SetFloatVector( lpv->minCorner );
	g_pLV_CellSize->SetFloat( lpv->cellSize );

	renderSceneFromCamera( pd3dDevice, g_ActiveCamera );

    DXUT_BeginPerfEvent( DXUT_PERFEVENTCOLOR, L"HUD / Stats" );
    RenderText();
    g_HUD.OnRender( fElapsedTime );
    g_SampleUI.OnRender( fElapsedTime );
    DXUT_EndPerfEvent();
}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10ResizedSwapChain 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10ReleasingSwapChain( void* pUserContext )
{
    g_DialogResourceManager.OnD3D10ReleasingSwapChain();
}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10CreateDevice 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10DestroyDevice( void* pUserContext )
{
	destroyScene();
	g_MeshLoader.Destroy();

    g_DialogResourceManager.OnD3D10DestroyDevice();
    g_SettingsDlg.OnD3D10DestroyDevice();
    SAFE_RELEASE( g_pFont10 );
    SAFE_RELEASE( g_pEffect10 );
    SAFE_RELEASE( g_pVertexLayout );
    SAFE_RELEASE( g_pSprite10 );

    SAFE_DELETE( g_pTxtHelper );
	SAFE_DELETE( lpv );

	SAFE_RELEASE( g_whiteDummyTexture );
	SAFE_RELEASE( g_whiteDummyTextureSRV );
}


//--------------------------------------------------------------------------------------
// Called right before creating a D3D9 or D3D10 device, allowing the app to modify the device settings as needed
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext )
{
    if( pDeviceSettings->ver == DXUT_D3D9_DEVICE )
    {
        IDirect3D9* pD3D = DXUTGetD3D9Object();
        D3DCAPS9 Caps;
        pD3D->GetDeviceCaps( pDeviceSettings->d3d9.AdapterOrdinal, pDeviceSettings->d3d9.DeviceType, &Caps );

        // If device doesn't support HW T&L or doesn't support 1.1 vertex shaders in HW 
        // then switch to SWVP.
        if( ( Caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT ) == 0 ||
            Caps.VertexShaderVersion < D3DVS_VERSION( 1, 1 ) )
        {
            pDeviceSettings->d3d9.BehaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
        }

        // Debugging vertex shaders requires either REF or software vertex processing 
        // and debugging pixel shaders requires REF.  
#ifdef DEBUG_VS
        if( pDeviceSettings->d3d9.DeviceType != D3DDEVTYPE_REF )
        {
            pDeviceSettings->d3d9.BehaviorFlags &= ~D3DCREATE_HARDWARE_VERTEXPROCESSING;
            pDeviceSettings->d3d9.BehaviorFlags &= ~D3DCREATE_PUREDEVICE;
            pDeviceSettings->d3d9.BehaviorFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;
        }
#endif
#ifdef DEBUG_PS
        pDeviceSettings->d3d9.DeviceType = D3DDEVTYPE_REF;
#endif
    }

    // For the first device created if its a REF device, optionally display a warning dialog box
    static bool s_bFirstTime = true;
    if( s_bFirstTime )
    {
        s_bFirstTime = false;
        if( ( DXUT_D3D9_DEVICE == pDeviceSettings->ver && pDeviceSettings->d3d9.DeviceType == D3DDEVTYPE_REF ) ||
            ( DXUT_D3D10_DEVICE == pDeviceSettings->ver &&
              pDeviceSettings->d3d10.DriverType == D3D10_DRIVER_TYPE_REFERENCE ) )
            DXUTDisplaySwitchingToREFWarning( pDeviceSettings->ver );
    }

    return true;
}


//--------------------------------------------------------------------------------------
// Handle updates to the scene.  This is called regardless of which D3D API is used
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext )
{
	// Update the camera's position based on user input 
    g_ActiveCamera->FrameMove( fElapsedTime );
}


//--------------------------------------------------------------------------------------
// Handle messages to the application
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing,
                          void* pUserContext )
{
    // Pass messages to dialog resource manager calls so GUI state is updated correctly
    *pbNoFurtherProcessing = g_DialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    // Pass messages to settings dialog if its active
    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
        return 0;
    }

    // Give the dialogs a chance to handle the message first
    *pbNoFurtherProcessing = g_HUD.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;
    *pbNoFurtherProcessing = g_SampleUI.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    // Pass all remaining windows messages to camera so it can respond to user input
    g_ActiveCamera->HandleMessages( hWnd, uMsg, wParam, lParam );

    return 0;
}


//--------------------------------------------------------------------------------------
// Handle key presses
//--------------------------------------------------------------------------------------
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
}


//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
    switch( nControlID )
    {
        case IDC_TOGGLEFULLSCREEN:
            DXUTToggleFullScreen(); break;
        case IDC_TOGGLEREF:
            DXUTToggleREF(); break;
        case IDC_TOGGLEWARP:
            DXUTToggleWARP(); break;
        case IDC_CHANGEDEVICE:
            g_SettingsDlg.SetActive( !g_SettingsDlg.IsActive() ); break;
		case IDC_SWITCH_CAMERA: {
			std::vector< CBaseCamera * >::const_iterator activeIter = std::find( g_RegisteredCameras.begin(), g_RegisteredCameras.end(), g_ActiveCamera );
			if( ++activeIter == g_RegisteredCameras.end() ) {
				g_ActiveCamera = g_RegisteredCameras.front();
			}
			else {
				g_ActiveCamera = *activeIter;
			}
			break;
		}
		case IDC_RELOAD_EFFECT:
			g_bReloadEffect = true; break;
		case IDC_DUMP:
			g_MeshLoader.Dump( gc_dumpName ); break;
		case IDC_RESTORE:
			g_bRestoreTriggered = true;
			break;
		case IDC_RELOAD_INJECTION_EFFECT:
			g_bReloadInjectionFX = true;
			break;
		case IDC_LPV_CHOOSER:
			g_iLPVIndex = g_SampleUI.GetSlider( IDC_LPV_CHOOSER )->GetValue();
			break;
		case IDC_OCCLUSION_SWITCH:
			lpv->withOcclusion = g_SampleUI.GetCheckBox( IDC_OCCLUSION_SWITCH )->GetChecked();
			break;
		case IDC_DIRECT_LIGHTING_SWITCH:
			g_bRenderDirectLighting = g_SampleUI.GetCheckBox( IDC_DIRECT_LIGHTING_SWITCH )->GetChecked();
			break;
		case IDC_GI_BOOST_VALUE:
			g_fIndirectLightingAttenuation = g_SampleUI.GetSlider( IDC_GI_BOOST_VALUE )->GetValue() / float( g_fIndirectLightingAttenuation_UnitSteps );
			g_SampleUI.GetStatic( IDC_GI_BOOST_INFO )->SetText( boost::lexical_cast<std::wstring>(g_fIndirectLightingAttenuation).c_str() );
			break;
    }
}


//////////////////////////////////////////////////////////////////////////
// scene stuff

// scene parts
struct Entity {
	D3DXVECTOR3 position;
	ID3DX10Mesh *mesh;
	Material material;

	void render( ID3D10Device *device, const RenderContext &renderContext ) {
		D3DXMATRIX localMatrix;
		D3DXMatrixTranslation( &localMatrix, position.x, position.y, position.z );

		D3DXMATRIX worldMatrix = localMatrix * renderContext.worldMatrix;

		Effect_setTransformation( worldMatrix, renderContext.viewMatrix, renderContext.projectionMatrix );
		device->IASetInputLayout( g_pVertexLayout );

		Effect_setMaterial( material );
		D3D10_TECHNIQUE_DESC techDesc;
		renderContext.technique->GetDesc( &techDesc );

		for ( UINT p = 0; p < techDesc.Passes; ++p ) {
			renderContext.technique->GetPassByIndex(p)->Apply(0);
			mesh->DrawSubset(0);
		}
	}

	Entity( const D3DXVECTOR3 &position, const Material &material, ID3DX10Mesh *mesh = NULL ) : mesh( mesh ), position( position ), material( material ) {
	}

	~Entity() {
		SAFE_RELEASE( mesh );
	}
};

Entity *box, *plane, *sphere, *volumeBox;

const D3DXVECTOR3 black = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
const D3DXVECTOR3 white = D3DXVECTOR3( 1.0f, 1.0f, 1.0f );
const D3DXVECTOR3 red = D3DXVECTOR3( 1.0f, 0.0f, 0.0f );
const D3DXVECTOR3 green = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );
const D3DXVECTOR3 blue = D3DXVECTOR3( 0.0f, 0.0f, 1.0f );

void initScene( ID3D10Device *device ) {
	HRESULT hr;

	plane = new Entity( D3DXVECTOR3( 0.0f, 0.0f, 0.0f ), Material( green ) );
	box = new Entity( D3DXVECTOR3( 3.0f, 0.5f + 1.5f, 3.0f ), Material( red ) );
	volumeBox = new Entity( D3DXVECTOR3( 0.0f, 5.0f, 0.0f ), Material( white ) );
	sphere = new Entity( D3DXVECTOR3( -2.0f, 0.5f + 2.0f, -2.0f ), Material( blue ) );

	V( DXUTCreateBox( device, 10.0f, 1.0f, 10.0f, &plane->mesh ) );
	V( DXUTCreateBox( device, 3.0f, 3.0f, 3.0f, &box->mesh ) );
	V( DXUTCreateBox( device, 1.0f, 1.0f, 1.0f, &volumeBox->mesh ) );
	V( DXUTCreateSphere( device, 2.0f, 10, 10, &sphere->mesh ) );

	g_Light.rsm.create( device, RSM_RESOLUTION );
	g_Light.sm.create( device, SHADOWMAP_RESOLUTION );

	D3DXVECTOR3 lightPos( 0.0f, 100.0f, 0.0f );
	D3DXVECTOR3 lightLookAt( 1.0f, 0.0, 1.0f );
	
	g_Light.camera.SetViewParams( &lightPos, &lightLookAt );
	g_Light.flux = white * 13 * 6000.0f;

	g_RegisteredCameras.push_back( &g_Light.camera );
}

void destroyScene() {
	g_Light.rsm.destroy();
	g_Light.sm.destroy();

	SAFE_DELETE( plane );
	SAFE_DELETE( box );
	SAFE_DELETE( sphere );
	SAFE_DELETE( volumeBox );
}

void renderLoadedMesh( ID3D10Device *device, const RenderContext &renderContext ) {
	D3DXMATRIX worldMatrix = renderContext.worldMatrix;

	Effect_setTransformation( worldMatrix, renderContext.viewMatrix, renderContext.projectionMatrix );
	
	UINT iCurSubset = ( UINT )( INT_PTR )g_SampleUI.GetComboBox( IDC_SUBSET )->GetSelectedData();

	//
	// Render the mesh
	//
	if ( iCurSubset == -1 )
	{
		for ( UINT iSubset = 0; iSubset < g_MeshLoader.GetNumSubsets(); ++iSubset )
		{
			renderSubset( renderContext, iSubset );
		}
	} else
	{
		renderSubset( renderContext, iCurSubset );
	}
}

void renderScene( ID3D10Device *device, const RenderContext &renderContext ) {
// 	if( !renderContext.rsmPass )
// 		Effect_setAmbientMaterial( 0.03f * white );
// 	else 
// 		Effect_setAmbientMaterial( black );

	plane->render( device, renderContext );
	sphere->render( device, renderContext );

	renderLoadedMesh( device, renderContext );

	box->render( device, renderContext );

	if( !renderContext.rsmPass ) {
		RenderContext subRenderContext = renderContext;
		subRenderContext.technique = g_pRenderRCTechnique;
		volumeBox->render( device, subRenderContext );
	}
	else
		volumeBox->render( device, renderContext );
}

void renderSceneFromCamera( ID3D10Device *device, const CBaseCamera *camera) {
	RenderContext renderContext;
	// Get the projection & view matrix from the camera class
	D3DXMatrixIdentity( &renderContext.worldMatrix );
	renderContext.viewMatrix = *camera->GetViewMatrix();
	renderContext.projectionMatrix = *camera->GetProjMatrix();
	renderContext.rsmPass = false;
	renderContext.technique = g_pRenderLVTechnique;

	// TODO [5/28/2010 Andreas]
	Effect_setCameraPosition( *camera->GetEyePt() );

	renderContext.activeLight = &g_Light;
	Effect_setLight( g_Light );

	renderScene( device, renderContext );
}

void renderSM( ID3D10Device *device, Light &light ) {
	D3D10_VIEWPORT oldViewport;
	UINT numViewports = 1;
	device->RSGetViewports( &numViewports, &oldViewport );

	RenderContext renderContext;

	renderContext.rsmPass = true;
	renderContext.activeLight = &light;
	renderContext.technique = g_pSMTechnique;

	// TODO: Effect_setRSM [6/17/2010 Andreas]
	g_pRSMResolution->SetInt( light.sm.resolution );

	D3DXMatrixPerspectiveFovLH( &renderContext.projectionMatrix, D3DX_PI / 2, 1.0, gc_zNear, gc_zFar );

	D3DXVECTOR3 up( 0.0f, 1.0f, 0.0f );
	D3DXMatrixLookAtLH( &renderContext.viewMatrix, &light.position(), &light.lookAt(), &up );
	D3DXMatrixIdentity( &renderContext.worldMatrix );

	light.worldViewProjMatrix = renderContext.worldMatrix * renderContext.viewMatrix * renderContext.projectionMatrix;

	light.sm.clearFace( device );
	light.sm.makeTarget( device );

	Effect_setLight( light, false );

	renderScene( device, renderContext );

	device->OMSetRenderTargets( 0, NULL, NULL );
	device->RSSetViewports( 1, &oldViewport );
}

void renderRSM( ID3D10Device *device, Light &light ) {
	D3D10_VIEWPORT oldViewport;
	UINT numViewports = 1;
	device->RSGetViewports( &numViewports, &oldViewport );

	RenderContext renderContext;

	renderContext.rsmPass = true;
	renderContext.activeLight = &light;
	renderContext.technique = g_pRSMTechnique;

	// TODO: Effect_setRSM [6/17/2010 Andreas]
	g_pRSMResolution->SetInt( light.rsm.resolution );

	D3DXMatrixPerspectiveFovLH( &renderContext.projectionMatrix, D3DX_PI / 2, 1.0, gc_zNear, gc_zFar );

	D3DXVECTOR3 up( 0.0f, 1.0f, 0.0f );
	D3DXMatrixLookAtLH( &renderContext.viewMatrix, &light.position(), &light.lookAt(), &up );
	D3DXMatrixIdentity( &renderContext.worldMatrix );

	light.worldViewProjMatrix = renderContext.worldMatrix * renderContext.viewMatrix * renderContext.projectionMatrix;
	

	light.rsm.clearFace( device );
	light.rsm.makeTarget( device );

	Effect_setLight( light, false );

	renderScene( device, renderContext );

	device->OMSetRenderTargets( 0, NULL, NULL );
	device->RSSetViewports( 1, &oldViewport );
}