//--------------------------------------------------------------------------------------
// File: scene.fx
//--------------------------------------------------------------------------------------
// File: lpv.fx
//
// (FreeBSD License)
// 
// Copyright 2010 Andreas Kirsch. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are
// permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice, this list of
// conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list
// of conditions and the following disclaimer in the documentation and/or other materials
// provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY ANDREAS KIRSCH ``AS IS'' AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ANDREAS KIRSCH OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
// ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// The views and conclusions contained in the software and documentation are those of the
// authors and should not be interpreted as representing official policies, either expressed
// or implied, of Andreas Kirsch.
//--------------------------------------------------------------------------------------

float4 SH_evaluate( float3 direction ) {
	const float band0Factor = 0.282094792f;
	const float band1Factor = 0.488602512f;
	
	return float4( band0Factor, -band1Factor * direction.y, band1Factor * direction.z, -band1Factor * direction.x );
}

//const float Pi = 3.1415926f;
#define Pi 3.1415926f

//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------
float3 g_vMaterialAmbient   = float3( 0.2f, 0.2f, 0.2f );   // Material's ambient color
float3 g_vMaterialDiffuse   = float3( 0.8f, 0.8f, 0.8f );   // Material's diffuse color
float3 g_vMaterialSpecular  = float3( 1.0f, 1.0f, 1.0f );   // Material's specular color
float  g_fMaterialAlpha     = 1.0f;
int    g_nMaterialShininess = 32;

float g_fLightAmbient		= 0.2f;

float3 g_vLightColor        = float3( 1.0f, 1.0f, 1.0f );     // Light color
float3 g_vLightPosition     = float3( 50.0f, 10.0f, 0.0f );   // Light position

int	g_RSMResolution;

float3 g_vCameraPosition;

Texture2D<float> g_LightDepthMap;
float4x4 g_mLightProjMatrix; // world->projected light space

Texture2D g_MeshTexture;        // Color texture for mesh

Texture3D<float4> g_LV[3]; // light volume
Texture3D<float4> g_GV; // occlusion volume

float3 g_LV_MinCorner;
float g_LV_CellSize;

float  g_fTime ;                // App's time in seconds
matrix g_mProjection;
matrix g_mWorld;
matrix g_mView;
matrix g_mInvWorld;
matrix g_mInvView;

matrix g_mWorldViewProjection ; // World * View * Projection matrix

bool g_renderDirectLighting;
float g_indirectLightingAttenuation;

DepthStencilState DepthState {
	DepthEnable = TRUE;
};

RasterizerState DefaultRS
{
	CullMode = NONE;
};

RasterizerState RasterizerRC
{
	DepthClipEnable = FALSE;
};

BlendState BlendRC
{
	BlendEnable[0] = true;
	SrcBlend = ONE;
	DestBlend = INV_SRC_ALPHA;
};

DepthStencilState EnableDepth
{
    DepthEnable = TRUE;
    DepthWriteMask = ALL;
};

DepthStencilState DisableDepth
{
    DepthEnable = FALSE;
    DepthWriteMask = ZERO;
};

BlendState NoBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = FALSE;
    RenderTargetWriteMask[0] = 0x0F;
};

BlendState AlphaBlendingPremultiplied
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = TRUE;
    SrcBlend = ONE;
    DestBlend = INV_SRC_ALPHA;
    BlendOp = ADD;
    SrcBlendAlpha = ZERO;
    DestBlendAlpha = ZERO;
    BlendOpAlpha = ADD;
    RenderTargetWriteMask[0] = 0x0F;
};

RasterizerState EnableCulling
{
    CullMode = BACK;
    MultisampleEnable = FALSE;
};

RasterizerState DisableCulling
{
    CullMode = NONE;
    MultisampleEnable = FALSE;    
};


//--------------------------------------------------------------------------------------
// Texture sampler
//--------------------------------------------------------------------------------------
SamplerState samLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
};

SamplerState  g_samShadow2D
{
	Filter = MIN_MAG_MIP_POINT;
};

SamplerComparisonState samplerShadow
{
	Filter = COMPARISON_MIN_MAG_LINEAR_MIP_POINT;
	ComparisonFunc = LESS;
    AddressU = BORDER;
    AddressV = BORDER;
    BorderColor = float4(0.0f, 0.0f, 0.0f, 0.0f);
};


SamplerState samLV
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Border;
    AddressV = Border;
    AddressW = Border;
    BorderColor = float4( 0,0,0,0 );
};

//--------------------------------------------------------------------------------------
// Vertex shader input structure
//--------------------------------------------------------------------------------------
struct VS_INPUT
{
    float3 vPosObject   : POSITION;
    float3 vNormalObject: NORMAL;
    float2 vTexUV       : TEXCOORD0;
};

//--------------------------------------------------------------------------------------
// Pixel shader input structure
//--------------------------------------------------------------------------------------
struct PS_INPUT
{
    float4 vPosProj : SV_POSITION;
    float4 vPosWorld : WORLD_POSITION;
    float3 vNormalWorld : WORLD_NORMAL;
    float2 vTexUV : TEXCOORD0;
    float sViewZ : VIEW_Z;
};

PS_INPUT VS_Render( VS_INPUT input )
{
    PS_INPUT output;
    
    // Transform the position into world space for lighting, and projected space
    // for display
    output.vPosWorld = mul( float4(input.vPosObject,1), g_mWorld );
    output.vPosProj = mul( float4(input.vPosObject,1), g_mWorldViewProjection );
	output.sViewZ = mul( output.vPosWorld, g_mView ).z;
    // Transform the normal into world space for lighting
    output.vNormalWorld = mul( input.vNormalObject, (float3x3)g_mWorld );

    // Pass the texture coordinate
    output.vTexUV = input.vTexUV;

    return output;
}

struct PS_OUTPUT {
	float4 vColor : SV_Target;
};

PS_OUTPUT PS_Render( PS_INPUT input )
{
    PS_OUTPUT output;
    
    // Compute the light vector
    float3 vLight = normalize( g_vLightPosition - input.vPosWorld.xyz );
    float sLightDistance = distance( g_vLightPosition, input.vPosWorld );

    // Compute the ambient and diffuse components of illumination
    output.vColor.rgb = g_vMaterialAmbient * g_fLightAmbient;
    output.vColor.rgb += g_vLightColor * g_vMaterialDiffuse / sLightDistance * saturate( dot( vLight, input.vNormalWorld ) );
    
    // Sample and modulate the texture
    output.vColor.rgb *= g_MeshTexture.Sample( samLinear, input.vTexUV );
    
    // calculate the specular term
    float3 vCamera = normalize( input.vPosWorld.xyz - g_vCameraPosition );
    float3 vReflection = reflect( vLight, input.vNormalWorld );
    float  fPhoneValue = saturate( dot(vReflection, vCamera) );

	//if( g_nMaterialShininess != 0.0 )
		output.vColor.rgb += g_vMaterialSpecular * pow( fPhoneValue, g_nMaterialShininess + 0.1 );

    output.vColor.a = 1.0f; //g_fMaterialAlpha;

    return output;
}

float evalShadowMap( float4 worldPos ) {
	float4 projPos = mul( worldPos, g_mLightProjMatrix );
	if( projPos.w < 0.1 ) {
		return 0.0f;
	}
	
	float3 homoPos = projPos.xyz / projPos.w;
	
	// homoPos.z is in 0..1
	
	float2 ShadowTexCoord = 0.5 * homoPos.xy + float2( 0.5, 0.5 );
	ShadowTexCoord.y = 1.0f - ShadowTexCoord.y;
 
	// transform to texel space
	const int SMAP_RESOLUTION = g_RSMResolution;
	const float SMAP_OFFSET = 1.0 / SMAP_RESOLUTION;
	const float SMAP_EPSILON = 0.00004f;

	float shadowMapDepth = 0.0;
		
	[unroll]
	for(float x = -1.5f; x <= 1.5f; x += 1.0f)
		[unroll]
		for(float y = -1.5f; y <= 1.5f; y += 1.0f)
		{
			shadowMapDepth += g_LightDepthMap.SampleCmpLevelZero(samplerShadow, ShadowTexCoord + float2( x, y ) * SMAP_OFFSET, homoPos.z - SMAP_EPSILON);
		}
			
	shadowMapDepth /= 16.0f;

	return shadowMapDepth;
}

#define RENDER_DIRECT_LIGHTING
#define RENDER_DIRECT_SHADOW
//#define RENDER_AMBIENT_LIGHTING
#define RENDER_INDIRECT_LIGHTING
PS_OUTPUT PS_RenderLV( PS_INPUT input ) {
    PS_OUTPUT output;
    
    // Compute the light vector
    float3 directRadiance;
    
#ifdef RENDER_DIRECT_LIGHTING
	if( g_renderDirectLighting ) {
		float3 vLight = normalize( g_vLightPosition - input.vPosWorld.xyz );
		float sLightDistance = distance( g_vLightPosition, input.vPosWorld );
		// g_vLightColor is the light flux
		float lightRadiance = g_vLightColor / (4 * Pi) / sLightDistance / sLightDistance;
		directRadiance = lightRadiance * saturate( dot( vLight, input.vNormalWorld ) ); 
  
#ifdef RENDER_DIRECT_SHADOW
		directRadiance *= evalShadowMap( input.vPosWorld );
#endif
	}
	else
#endif
	{
		directRadiance = float3( 0, 0, 0 );
	}
    
    int3 LVsize;
    g_LV[0].GetDimensions( LVsize.x, LVsize.y, LVsize.z );
    
    // Compute the light cell coords
    //int4 lightCellCoords = int4( int3( (input.vPosWorld - g_LV_MinCorner) / g_LV_CellSize + 0.5 * input.vNormalWorld ), 0 );

    float4 SHintensity = SH_evaluate( -input.vNormalWorld );
    float3 lpvCellCoords = (input.vPosWorld - g_LV_MinCorner) / g_LV_CellSize / LVsize;
    float3 lpvIntensity = float3( 
						dot( SHintensity, g_LV[0].Sample( samLV, lpvCellCoords ) ),
						dot( SHintensity, g_LV[1].Sample( samLV, lpvCellCoords ) ),
						dot( SHintensity, g_LV[2].Sample( samLV, lpvCellCoords ) )
					);

    // Compute the ambient and diffuse components of illumination
#ifdef RENDER_AMBIENT_LIGHTING
    output.vColor.rgb = g_vMaterialAmbient * g_fLightAmbient;
#else
	output.vColor.rgb = float3( 0.0, 0.0, 0.0 );
#endif

#ifdef RENDER_INDIRECT_LIGHTING
	float3 lpvRadiance = g_indirectLightingAttenuation * max( lpvIntensity * 4 / g_LV_CellSize / g_LV_CellSize, 0 );
#else
	float3 lpvRadiance = float3( 0.0, 0.0, 0.0 );
#endif
	
    output.vColor.rgb += (lpvRadiance + directRadiance) * g_vMaterialDiffuse;
    // Sample and modulate the texture
    output.vColor.rgb *= g_MeshTexture.Sample( samLinear, input.vTexUV );
    
    
    return output;
}

struct PS_RC_INPUT
{
    float4 vPosProj : SV_POSITION;
    float3 vPosObject : OBJECT_POSITION;
};

PS_RC_INPUT VS_RenderRC( VS_INPUT input ) {
    PS_RC_INPUT output;
    
    // Transform the position into world space for lighting, and projected space
    // for display
    output.vPosObject = input.vPosObject;
    output.vPosProj = mul( float4(input.vPosObject,1), g_mWorldViewProjection );
	
    return output;
}

//#define RC_FIRST_HIT_BREAK
#define RC_LPV
//#define RC_DIRECTION_DEPENDENT
PS_OUTPUT PS_RenderRC( PS_RC_INPUT input )  {
	PS_OUTPUT output;
	
	float3 localOrigin = mul( mul( float4( 0.0, 0.0, 0.0, 1.0 ), g_mInvView ), g_mInvWorld );
	
	float3 direction = normalize( input.vPosObject - localOrigin );
	float4 SHintensity = SH_evaluate( -direction );
	
	int3 LVsize;
	g_LV[0].GetDimensions( LVsize.x, LVsize.y, LVsize.z );

	const int maxIterations=512;
	float3 volumePosition = (input.vPosObject + 0.5) * LVsize;
	
	//output.vColor = float4(input.vViewDirectionObject / 2 + 0.5, 1.0);
	//return output;

	float4 rayValue = float4( 0, 0, 0, 0.0 );
	int i;
	for( i = 0 ; i < maxIterations ; i++ ) {
		if( any( volumePosition < 0.0 ) || any( volumePosition > LVsize ) ) {
			break;
		}

		int4 intPos = int4( volumePosition, 0 );
#ifdef RC_LPV
// LPV
		float3 lightIntensity =
#ifdef RC_DIRECTION_DEPENDENT
					float3( 
						dot( SHintensity, g_LV[0].Load( intPos ) ),
						dot( SHintensity, g_LV[1].Load( intPos ) ),
						dot( SHintensity, g_LV[2].Load( intPos ) )
					);
#else
					float3( 
						dot( float4( 1.0, 1.0, 1.0, 1.0 ), abs( g_LV[0].Load( intPos ) )),
						dot( float4( 1.0, 1.0, 1.0, 1.0 ), abs( g_LV[1].Load( intPos ) )),
						dot( float4( 1.0, 1.0, 1.0, 1.0 ), abs( g_LV[2].Load( intPos ) ))
					);
#endif


		float monoIntensity = (lightIntensity[0]+lightIntensity[1]+lightIntensity[2]) / 3;
		//rayValue.rgb = max( rayValue.rgb, lightIntensity.rgb );
		rayValue.rgb += (1 - rayValue.a) * lightIntensity * 4 / g_LV_CellSize / g_LV_CellSize;
		rayValue.a += (1 - rayValue.a) * monoIntensity;
#else
// OCV
		float monoIntensity =
#ifdef RC_DIRECTION_DEPENDENT
					dot( SHintensity, g_GV.Load( intPos ) );
#else
					dot( float4( 1, 1, 1, 1 ), abs( g_GV.Load( intPos ) ) );
#endif
		rayValue.rgb += (1 - rayValue.a) * monoIntensity;
		rayValue.a += (1 - rayValue.a) * monoIntensity;
		
#endif

#ifdef RC_FIRST_HIT_BREAK
		if( monoIntensity > 0 ) {
			break;
		}
#endif

		if( rayValue.a >= 1 ) {
			break;
		}

		volumePosition += direction;
	}
	
	//output.vColor.rgb = i / 100;
	rayValue.a = max( rayValue, 0.5 );
	output.vColor = rayValue;
	return output;
}

struct PS_RSM_OUTPUT {
	float3 vFlux : SV_Target0;
	float3 vNormalWorld : SV_Target1;
	// store light z in vPosWorld.w for now!
	float4 vPosWorld : SV_Target2;
};

PS_RSM_OUTPUT PS_RSMRender( PS_INPUT input )
{
    PS_RSM_OUTPUT output;
    
    // Compute the light vector
    float3 vLight = g_vLightPosition - input.vPosWorld.xyz;
    float3 vLightDir = normalize( vLight );

	// fov of 90 with aspect ratio of 1:1 => total subtended solid angle of 4Pi / 6 = 2.09439510f
	//float subtendedSolidAnglePerTexel = 2.09439510f / g_RSMResolution / g_RSMResolution;
	float fluxQuota = 1.0 / 6.0 / g_RSMResolution / g_RSMResolution;

    // Compute the diffuse components of illumination
    // g_vLightColor == total light flux
    output.vFlux = g_vLightColor * fluxQuota * g_vMaterialDiffuse * g_MeshTexture.Sample( samLinear, input.vTexUV ) * saturate( dot( vLightDir, input.vNormalWorld ) );
    
    output.vNormalWorld = input.vNormalWorld;
    output.vPosWorld.xyz = input.vPosWorld.xyz;    
    //output.vColor.rgb = vNormalWorld * 0.5 + 0.5;
    output.vPosWorld.w = input.sViewZ;

    return output;
}

float4 PS_SMRender( PS_INPUT input ) : SV_Target
{
	return float4( 1.0, 1.0, 1.0, 1.0 );
}

//--------------------------------------------------------------------------------------
// Techniques
//--------------------------------------------------------------------------------------

technique10 UnifiedRendering
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS_Render() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_Render() ) );
        
        SetDepthStencilState( DepthState, 0 );
    }
}

technique10 UnifiedLVRendering
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS_Render() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_RenderLV() ) );
        
        SetDepthStencilState( DepthState, 0 );
    }
}

technique10 UnifiedRCRendering
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS_RenderRC() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_RenderRC() ) );
        
        SetDepthStencilState( DepthState, 0 );
        SetBlendState( BlendRC, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetRasterizerState( RasterizerRC );
    }
}

technique10 RSMRendering
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS_Render() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_RSMRender() ) );
        
        SetBlendState(NoBlending, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xFFFFFFFF);
        SetDepthStencilState(EnableDepth, 0);
        SetRasterizerState(EnableCulling);
    }
}

technique10 SMRendering
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS_Render() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_SMRender() ) );
        //SetPixelShader( NULL );
        
        SetBlendState(NoBlending, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xFFFFFFFF);
        SetDepthStencilState(EnableDepth, 0);
        SetRasterizerState(EnableCulling);
    }
}